import time


class Stopwatch:
    """
        Simple time measuring class.
    """
    def __init__(self):
        self.timestamp = None
        self.time = None

    def start(self):
        """
            Start a time measurement
        """
        self.time = None
        self.timestamp = time.time()

    def lap(self):
        cur_time = time.time()
        start_time = self.timestamp
        self.timestamp = cur_time
        return cur_time - start_time

    @property
    def delta(self):
        if not self.timestamp:
            # stopped or not started
            return self.time or 0
        return time.time() - self.timestamp

    def stop(self):
        """
            Stop a running time measurement. returns the measured time
        """
        cur_time = time.time()
        start_time = self.timestamp
        self.time = cur_time - start_time
        self.timestamp = None
        return self.time

    def __enter__(self):
        """
            Start stopwatch when entering with-block.
        """
        self.start()

    def __exit__(self, tpe, value, tb):
        """
            Stop stopwatch when exiting with-block.
        """
        self.stop()

    def __str__(self):
        cur_time = time.time()
        if self.timestamp:
            return str(self.time or cur_time - self.timestamp)
        else:
            return 'Not started'




import socket
import requests


def get_ip():
    # taken from https://stackoverflow.com/a/28950776/13703283
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


def get_external_ip():
    return requests.get('https://api.ipify.org').text


def is_connected_dns(host="1.1.1.1"):
    try:
        # connect to the host -- tells us if the host is actually reachable
        socket.create_connection((host, 53))
        return True
    except OSError:
        pass
    return False


def is_connected_http(host="http://duckduckgo.com"):
    try:
        status = requests.head(host).status_code
        return True
    except OSError:
        pass
    return False


def is_connected():
    return any((is_connected_dns(), is_connected_http()))



from os.path import exists, isfile

from adapt.intent import Intent
from mycroft_bus_client import MessageBusClient
from mycroft_bus_client.message import Message, dig_for_message
from ovos_utils import create_daemon
from ovos_utils.log import LOG


def to_alnum(skill_id):
    """Convert a skill id to only alphanumeric characters

     Non alpha-numeric characters are converted to "_"

    Args:
        skill_id (str): identifier to be converted
    Returns:
        (str) String of letters
    """
    return ''.join(c if c.isalnum() else '_' for c in str(skill_id))


def munge_regex(regex, skill_id):
    """Insert skill id as letters into match groups.

    Args:
        regex (str): regex string
        skill_id (str): skill identifier
    Returns:
        (str) munged regex
    """
    base = '(?P<' + to_alnum(skill_id)
    return base.join(regex.split('(?P<'))


def munge_intent_parser(intent_parser, name, skill_id):
    """Rename intent keywords to make them skill exclusive
    This gives the intent parser an exclusive name in the
    format <skill_id>:<name>.  The keywords are given unique
    names in the format <Skill id as letters><Intent name>.

    The function will not munge instances that's already been
    munged

    Args:
        intent_parser: (IntentParser) object to update
        name: (str) Skill name
        skill_id: (int) skill identifier
    """
    # Munge parser name
    if not name.startswith(str(skill_id) + ':'):
        intent_parser.name = str(skill_id) + ':' + name
    else:
        intent_parser.name = name

    # Munge keywords
    skill_id = to_alnum(skill_id)
    # Munge required keyword
    reqs = []
    for i in intent_parser.requires:
        if not i[0].startswith(skill_id):
            kw = (skill_id + i[0], skill_id + i[0])
            reqs.append(kw)
        else:
            reqs.append(i)
    intent_parser.requires = reqs

    # Munge optional keywords
    opts = []
    for i in intent_parser.optional:
        if not i[0].startswith(skill_id):
            kw = (skill_id + i[0], skill_id + i[0])
            opts.append(kw)
        else:
            opts.append(i)
    intent_parser.optional = opts

    # Munge at_least_one keywords
    at_least_one = []
    for i in intent_parser.at_least_one:
        element = [skill_id + e.replace(skill_id, '') for e in i]
        at_least_one.append(tuple(element))
    intent_parser.at_least_one = at_least_one


class IntentServiceInterface:
    """Interface to communicate with the Mycroft intent service.

    This class wraps the messagebus interface of the intent service allowing
    for easier interaction with the service. It wraps both the Adapt and
    Padatious parts of the intent services.
    """

    def __init__(self, bus=None):
        self.bus = bus
        self.skill_id = self.__class__.__name__
        self.registered_intents = []
        self.detached_intents = []

    def set_bus(self, bus):
        self.bus = bus

    def set_id(self, skill_id):
        self.skill_id = skill_id

    def register_adapt_keyword(self, vocab_type, entity, aliases=None,
                               lang=None):
        """Send a message to the intent service to add an Adapt keyword.

            vocab_type(str): Keyword reference
            entity (str): Primary keyword
            aliases (list): List of alternative kewords
        """
        aliases = aliases or []
        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg.forward("register_vocab",
                                  {'start': entity,
                                   'end': vocab_type,
                                   'lang': lang}))
        for alias in aliases:
            self.bus.emit(msg.forward("register_vocab",
                                      {'start': alias,
                                       'end': vocab_type,
                                       'alias_of': entity,
                                       'lang': lang}))

    def register_adapt_regex(self, regex, lang=None):
        """Register a regex with the intent service.

        Args:
            regex (str): Regex to be registered, (Adapt extracts keyword
                         reference from named match group.
        """
        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg.forward("register_vocab",
                                  {'regex': regex, 'lang': lang}))

    def register_adapt_intent(self, name, intent_parser):
        """Register an Adapt intent parser object.

        Serializes the intent_parser and sends it over the messagebus to
        registered.
        """
        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg.forward("register_intent", intent_parser.__dict__))
        self.registered_intents.append((name, intent_parser))
        self.detached_intents = [detached for detached in self.detached_intents
                                 if detached[0] != name]

    def detach_intent(self, intent_name):
        """Remove an intent from the intent service.

        The intent is saved in the list of detached intents for use when
        re-enabling an intent.

        Args:
            intent_name(str): Intent reference
        """
        # split skill_id if needed
        if intent_name not in self and ":" in intent_name:
            name = intent_name.split(':')[1]
        else:
            name = intent_name

        if name in self:
            msg = dig_for_message() or Message("")
            if "skill_id" not in msg.context:
                msg.context["skill_id"] = self.skill_id
            self.bus.emit(msg.forward("detach_intent",
                                      {"intent_name": intent_name}))
            self.detached_intents.append((name, self.get_intent(name)))
            self.registered_intents = [pair for pair in self.registered_intents
                                       if pair[0] != name]

    def set_adapt_context(self, context, word, origin):
        """Set an Adapt context.

        Args:
            context (str): context keyword name
            word (str): word to register
            origin (str): original origin of the context (for cross context)
        """
        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg.forward('add_context',
                                  {'context': context, 'word': word,
                                   'origin': origin}))

    def remove_adapt_context(self, context):
        """Remove an active Adapt context.

        Args:
            context(str): name of context to remove
        """
        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg.forward('remove_context', {'context': context}))

    def register_padatious_intent(self, intent_name, filename, lang):
        """Register a padatious intent file with Padatious.

        Args:
            intent_name(str): intent identifier
            filename(str): complete file path for entity file
        """
        if not isinstance(filename, str):
            raise ValueError('Filename path must be a string')
        if not exists(filename):
            raise FileNotFoundError('Unable to find "{}"'.format(filename))

        data = {'file_name': filename,
                'name': intent_name,
                'lang': lang}
        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg.forward("padatious:register_intent", data))
        self.registered_intents.append((intent_name.split(':')[-1], data))

    def register_padatious_entity(self, entity_name, filename, lang):
        """Register a padatious entity file with Padatious.

        Args:
            entity_name(str): entity name
            filename(str): complete file path for entity file
        """
        if not isinstance(filename, str):
            raise ValueError('Filename path must be a string')
        if not exists(filename):
            raise FileNotFoundError('Unable to find "{}"'.format(filename))
        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg.forward('padatious:register_entity',
                                  {'file_name': filename,
                                   'name': entity_name,
                                   'lang': lang}))

    def __iter__(self):
        """Iterator over the registered intents.

        Returns an iterator returning name-handler pairs of the registered
        intent handlers.
        """
        return iter(self.registered_intents)

    def __contains__(self, val):
        """Checks if an intent name has been registered."""
        return val in [i[0] for i in self.registered_intents]

    def get_intent_names(self):
        return [i[0] for i in self.registered_intents]

    def detach_all(self):
        for name in self.get_intent_names():
            self.detach_intent(name)
        self.registered_intents = []
        self.detached_intents = []

    def get_intent(self, intent_name):
        """Get intent from intent_name.

        This will find both enabled and disabled intents.

        Args:
            intent_name (str): name to find.

        Returns:
            Found intent or None if none were found.
        """
        for name, intent in self:
            if name == intent_name:
                return intent
        for name, intent in self.detached_intents:
            if name == intent_name:
                return intent
        return None


class IntentQueryApi:
    """
    Query Intent Service at runtime
    """

    def __init__(self, bus=None, timeout=5):
        if bus is None:
            bus = MessageBusClient()
            create_daemon(bus.run_forever)
        self.bus = bus
        self.timeout = timeout

    def get_adapt_intent(self, utterance, lang="en-us"):
        """ get best adapt intent for utterance """
        msg = Message("intent.service.adapt.get",
                      {"utterance": utterance, "lang": lang},
                      context={"destination": "intent_service",
                               "source": "intent_api"})

        resp = self.bus.wait_for_response(msg,
                                          'intent.service.adapt.reply',
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None
        return data["intent"]

    def get_padatious_intent(self, utterance, lang="en-us"):
        """ get best padatious intent for utterance """
        msg = Message("intent.service.padatious.get",
                      {"utterance": utterance, "lang": lang},
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        resp = self.bus.wait_for_response(msg,
                                          'intent.service.padatious.reply',
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None
        return data["intent"]

    def get_intent(self, utterance, lang="en-us"):
        """ get best intent for utterance """
        msg = Message("intent.service.intent.get",
                      {"utterance": utterance, "lang": lang},
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        resp = self.bus.wait_for_response(msg,
                                          'intent.service.intent.reply',
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None
        return data["intent"]

    def get_skill(self, utterance, lang="en-us"):
        """ get skill that utterance will trigger """
        intent = self.get_intent(utterance, lang)
        if not intent:
            return None
        # theoretically skill_id might be missing
        if intent.get("skill_id"):
            return intent["skill_id"]
        # retrieve skill from munged intent name
        if intent.get("intent_name"):  # padatious + adapt
            return intent["name"].split(":")[0]
        if intent.get("intent_type"):  # adapt
            return intent["intent_type"].split(":")[0]
        return None  # raise some error here maybe? this should never happen

    def get_skills_manifest(self):
        msg = Message("intent.service.skills.get",
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        resp = self.bus.wait_for_response(msg,
                                          'intent.service.skills.reply',
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None
        return data["skills"]

    def get_active_skills(self, include_timestamps=False):
        msg = Message("intent.service.active_skills.get",
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        resp = self.bus.wait_for_response(msg,
                                          'intent.service.active_skills.reply',
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None
        if include_timestamps:
            return data["skills"]
        return [s[0] for s in data["skills"]]

    def get_adapt_manifest(self):
        msg = Message("intent.service.adapt.manifest.get",
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        resp = self.bus.wait_for_response(msg,
                                          'intent.service.adapt.manifest',
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None
        return data["intents"]

    def get_padatious_manifest(self):
        msg = Message("intent.service.padatious.manifest.get",
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        resp = self.bus.wait_for_response(msg,
                                          'intent.service.padatious.manifest',
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None
        return data["intents"]

    def get_intent_manifest(self):
        padatious = self.get_padatious_manifest()
        adapt = self.get_adapt_manifest()
        return {"adapt": adapt,
                "padatious": padatious}

    def get_vocab_manifest(self):
        msg = Message("intent.service.adapt.vocab.manifest.get",
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        reply_msg_type = 'intent.service.adapt.vocab.manifest'
        resp = self.bus.wait_for_response(msg,
                                          reply_msg_type,
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None

        vocab = {}
        for voc in data["vocab"]:
            if voc.get("regex"):
                continue
            if voc["end"] not in vocab:
                vocab[voc["end"]] = {"samples": []}
            vocab[voc["end"]]["samples"].append(voc["start"])
        return [{"name": voc, "samples": vocab[voc]["samples"]}
                for voc in vocab]

    def get_regex_manifest(self):
        msg = Message("intent.service.adapt.vocab.manifest.get",
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        reply_msg_type = 'intent.service.adapt.vocab.manifest'
        resp = self.bus.wait_for_response(msg,
                                          reply_msg_type,
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None

        vocab = {}
        for voc in data["vocab"]:
            if not voc.get("regex"):
                continue
            name = voc["regex"].split("(?P<")[-1].split(">")[0]
            if name not in vocab:
                vocab[name] = {"samples": []}
            vocab[name]["samples"].append(voc["regex"])
        return [{"name": voc, "regexes": vocab[voc]["samples"]}
                for voc in vocab]

    def get_entities_manifest(self):
        msg = Message("intent.service.padatious.entities.manifest.get",
                      context={"destination": "intent_service",
                               "source": "intent_api"})
        reply_msg_type = 'intent.service.padatious.entities.manifest'
        resp = self.bus.wait_for_response(msg,
                                          reply_msg_type,
                                          timeout=self.timeout)
        data = resp.data if resp is not None else {}
        if not data:
            LOG.error("Intent Service timed out!")
            return None

        entities = []
        # read files
        for ent in data["entities"]:
            if isfile(ent["file_name"]):
                with open(ent["file_name"]) as f:
                    lines = f.read().replace("(", "").replace(")", "").split(
                        "\n")
                samples = []
                for l in lines:
                    samples += [a.strip() for a in l.split("|") if a.strip()]
                entities.append({"name": ent["name"], "samples": samples})
        return entities

    def get_keywords_manifest(self):
        padatious = self.get_entities_manifest()
        adapt = self.get_vocab_manifest()
        regex = self.get_regex_manifest()
        return {"adapt": adapt,
                "padatious": padatious,
                "regex": regex}


def open_intent_envelope(message):
    """Convert dictionary received over messagebus to Intent."""
    intent_dict = message.data
    return Intent(intent_dict.get('name'),
                  intent_dict.get('requires'),
                  intent_dict.get('at_least_one'),
                  intent_dict.get('optional'))



try:
    from phoneme_guesser import guess_phonemes as _guess_phonemes, \
        get_phonemes as _get_phonemes


    def guess_phonemes(word, lang="en-us"):
        return _guess_phonemes(word, lang)


    def get_phonemes(name, lang="en-us"):
        return _get_phonemes(name, lang)

except ImportError:
    from ovos_utils.log import LOG

    LOG.warning("phoneme_guesser optional import is not available")

    def guess_phonemes(word, lang="en-us"):
        raise ImportError("pip install phoneme_guesser")

    def get_phonemes(name, lang="en-us"):
        raise ImportError("pip install phoneme_guesser")

########################################################################
# ARPABET was invented for English.
# The standard dictionary written in ARPABET is the CMU dictionary.

arpabet2ipa = {
    'AA': 'ɑ',
    'AE': 'æ',
    'AH': 'ʌ',
    'AH0': 'ə',
    'AO': 'ɔ',
    'AW': 'aʊ',
    'AY': 'aɪ',
    'EH': 'ɛ',
    'ER': 'ɝ',
    'ER0': 'ɚ',
    'EY': 'eɪ',
    'IH': 'ɪ',
    'IH0': 'ɨ',
    'IY': 'i',
    'OW': 'oʊ',
    'OY': 'ɔɪ',
    'UH': 'ʊ',
    'UW': 'u',
    'B': 'b',
    'CH': 'tʃ',
    'D': 'd',
    'DH': 'ð',
    'EL': 'l̩ ',
    'EM': 'm̩',
    'EN': 'n̩',
    'F': 'f',
    'G': 'ɡ',
    'HH': 'h',
    'JH': 'dʒ',
    'K': 'k',
    'L': 'l',
    'M': 'm',
    'N': 'n',
    'NG': 'ŋ',
    'P': 'p',
    'Q': 'ʔ',
    'R': 'ɹ',
    'S': 's',
    'SH': 'ʃ',
    'T': 't',
    'TH': 'θ',
    'V': 'v',
    'W': 'w',
    'WH': 'ʍ',
    'Y': 'j',
    'Z': 'z',
    'ZH': 'ʒ'
}

ipa2arpabet = {v: k for k, v in arpabet2ipa.items()}



import unittest
from ovos_utils import rotate_list, camel_case_split, get_handler_name


class TestHelpers(unittest.TestCase):

    def test_utils(self):
        def some_function():
            return

        self.assertEqual(get_handler_name(some_function), "some_function")
        self.assertEqual(get_handler_name(self.test_utils), "test_utils")

        self.assertEqual(camel_case_split("MyAwesomeSkill"),
                         "My Awesome Skill")

        self.assertEqual(rotate_list([1, 2, 3]), [2, 3, 1])
        self.assertEqual(rotate_list([1, 2, 3], 2), [3, 1, 2])
        self.assertEqual(rotate_list([1, 2, 3], 3), [1, 2, 3])

        self.assertEqual(rotate_list([1, 2, 3], -1), [3, 1, 2])
        self.assertEqual(rotate_list([1, 2, 3], -2), [2, 3, 1])
        self.assertEqual(rotate_list([1, 2, 3], -3), [1, 2, 3])



from difflib import SequenceMatcher
import re
from enum import IntEnum, auto
from ovos_utils.log import LOG

try:
    import rapidfuzz
except ImportError:
    rapidfuzz = None


class MatchStrategy(IntEnum):
    SIMPLE_RATIO = auto()
    RATIO = auto()
    PARTIAL_RATIO = auto()
    TOKEN_SORT_RATIO = auto()
    TOKEN_SET_RATIO = auto()
    PARTIAL_TOKEN_RATIO = auto()
    PARTIAL_TOKEN_SORT_RATIO = auto()
    PARTIAL_TOKEN_SET_RATIO = auto()


def _validate_matching_strategy(strategy):
    if rapidfuzz is None and strategy != MatchStrategy.SIMPLE_RATIO:
        LOG.error("rapidfuzz is not installed, "
                  "falling back to SequenceMatcher for all match strategies")
        LOG.warning("pip install rapidfuzz")
        return MatchStrategy.SIMPLE_RATIO
    return strategy


def fuzzy_match(x, against, strategy=MatchStrategy.SIMPLE_RATIO):
    """Perform a 'fuzzy' comparison between two strings.
    Returns:
        float: match percentage -- 1.0 for perfect match,
               down to 0.0 for no match at all.
    """
    strategy = _validate_matching_strategy(strategy)
    if strategy == MatchStrategy.RATIO:
        score = rapidfuzz.fuzz.ratio(x, against) / 100
    elif strategy == MatchStrategy.PARTIAL_RATIO:
        score = rapidfuzz.fuzz.partial_ratio(x, against) / 100
    elif strategy == MatchStrategy.TOKEN_SORT_RATIO:
        score = rapidfuzz.fuzz.token_sort_ratio(x, against) / 100
    elif strategy == MatchStrategy.TOKEN_SET_RATIO:
        score = rapidfuzz.fuzz.token_set_ratio(x, against) / 100
    elif strategy == MatchStrategy.PARTIAL_TOKEN_SORT_RATIO:
        score = rapidfuzz.fuzz.partial_token_sort_ratio(x, against) / 100
    elif strategy == MatchStrategy.PARTIAL_TOKEN_SET_RATIO:
        score = rapidfuzz.fuzz.partial_token_set_ratio(x, against) / 100
    elif strategy == MatchStrategy.PARTIAL_TOKEN_RATIO:
        score = rapidfuzz.fuzz.partial_token_ratio(x, against) / 100
    else:
        score = SequenceMatcher(None, x, against).ratio()

    return score


def match_one(query, choices, match_func=None, strategy=MatchStrategy.SIMPLE_RATIO):
    """
        Find best match from a list or dictionary given an input

        Arguments:
            query:   string to test
            choices: list or dictionary of choices

        Returns: tuple with best match, score
    """
    return match_all(query, choices, match_func, strategy)[0]


def match_all(query, choices, match_func=None, strategy=MatchStrategy.SIMPLE_RATIO):
    """
        match scores from a list or dictionary given an input

        Arguments:
            query:   string to test
            choices: list or dictionary of choices

        Returns: list of tuples (match, score)
    """
    strategy = _validate_matching_strategy(strategy)
    match_func = match_func or fuzzy_match
    if isinstance(choices, dict):
        _choices = list(choices.keys())
    elif isinstance(choices, list):
        _choices = choices
    else:
        raise ValueError('a list or dict of choices must be provided')
    matches = []
    for c in _choices:
        if isinstance(choices, dict):
            matches.append((choices[c], match_func(query, c, strategy)))
        else:
            matches.append((c, match_func(query, c, strategy)))

    # TODO solve ties

    return sorted(matches, key=lambda k: k[1], reverse=True)


def singularize(word, lang="en"):
    return word.rstrip("s")


def split_sentences(text, new_lines=False):
    if new_lines:
        return text.split("\n")

    # normalize ambiguous cases
    words = text.split(" ")
    for idx, w in enumerate(words):
        # prev_w = words[idx-1] if idx > 0 else ""
        # next_w = words[idx + 1] if idx < len(words) - 1 else ""
        if w == ".":
            # handled ambiguous cases
            # "hello . He said" # regex handles these next

            # ignored ambiguous cases
            # "hello . he said"
            pass
        elif "." in w:
            # ignored ambiguous cases
            # "hello. he said"  # could be "Jones Jr. thinks ..."
            # "hello.he said"  # could be  "www.hello.com"
            # "hellO.He said"  # could be  "A.E.I.O.U"

            # handled cases
            # "hello.He said"
            if len(w.split(".")) == 2:
                prev_w, next_w = w.split(".")
                if prev_w and next_w and prev_w[-1].islower() and \
                        next_w[0].isupper():
                    words[idx] = w.replace(".", ";")
    text = " ".join(words)
    # ignored ambiguous cases
    # "hello. he said"  # could be "Jones Jr. thinks ..."
    #

    # handle punctuation delimiters except .
    delims = ["\n", ";", "!", "?"]
    _sentences = [s.strip() for s in re.split(r'(!|\?|\;|\n)*', text) if s not in
                  delims and s.strip()]

    sentences = []
    # handle . but be careful with numbers / names / websites?
    for sent in _sentences:
        sentences += [s.strip() for s in
                      re.split(r'(?<=[^A-Z].[.]) +(?=[A-Z])', sent) if
                      s.strip()]

    return sentences


def remove_parentheses(answer):
    # remove [xx] (xx) {xx}
    answer = re.sub(r'\[[^)]*\]', '', answer)
    answer = re.sub(r'\([^)]*\)', '', answer)
    answer = re.sub(r'\{[^)]*\}', '', answer)
    answer = answer.replace("(", "").replace(")", "") \
        .replace("[", "").replace("]", "").replace("{", "") \
        .replace("}", "").strip()
    # remove extra spaces
    words = [w for w in answer.split(" ") if w.strip()]
    answer = " ".join(words)
    if not answer:
        return None
    return answer


def summarize(answer):
    if not answer:
        return None
    return remove_parentheses(split_sentences(answer)[0])


def search_in_text(query, text, lang="en", all_matches=False,
                   thresh=0.1, paragraphs=True):
    words = query.split(" ")
    for idx, word in enumerate(words):
        words[idx] = singularize(word).lower().strip()
    query = " ".join(words)

    # search text inside sections
    candidates = split_sentences(text, paragraphs)
    scores = [0 for i in range(len(candidates))]
    for idx, c in enumerate(candidates):
        if len(c.split(" ")) < 4:
            continue
        score = scores[idx]
        for word in c.split():
            # total half assed scoring metric #1
            # each time query appears in sentence/paragraph boost score
            if len(word) < 3:
                continue
            word = word.lower().strip()
            singular_word = singularize(word, lang).lower().strip()

            if query == word:
                # magic numbers are bad
                score += 0.4
            elif query == singular_word:
                # magic numbers are bad
                score += 0.3
            # partial match
            elif query in word or word in query:
                score += 0.15
            elif query in singular_word or singular_word in query:
                score += 0.1

        scores[idx] = score

    # total half assed scoring metric #3
    # give preference to short sentences
    if not paragraphs:
        for idx, c in enumerate(candidates):
            scores[idx] = scores[idx] / (len(c) / 200 + 0.3)
    else:
        for idx, c in enumerate(candidates):
            scores[idx] = scores[idx] / (len(split_sentences(c)) + 0.1)

    best_score = max(scores)

    # this is a fake percent, sorry folks
    if best_score > 1:
        dif = best_score - 1
        scores = [s - dif for s in scores]
        scores = [s if s > 0 else 0.0001 for s in scores]
        best_score = 1

    if not all_matches:
        best = candidates[scores.index(best_score)]
        return best, best_score

    data = []
    for idx, c in enumerate(candidates):
        if c.strip() and scores[idx] >= thresh:
            data.append((c, scores[idx]))
    data.sort(key=lambda k: k[1], reverse=True)
    return data


def extract_sentences(query, text, lang="en"):
    return search_in_text(query, text, lang,
                          all_matches=True, paragraphs=False)


def extract_paragraphs(query, text, lang="en"):
    return search_in_text(query, text, lang,
                          all_matches=True, paragraphs=True)




from ovos_utils.enclosure.mark1.faceplate import FacePlateAnimation
import copy
import random


# Game of Life Base
class GoL(FacePlateAnimation):

    def __init__(self, entropy=0, grid=None, bus=None):
        super().__init__(grid, bus)
        self.entropy = entropy
        if self.is_empty:
            self.randomize()

    def _live_neighbours(self, y, x):
        """Returns the number of live neighbours."""
        count = 0
        if y > 0:
            if self.grid[y - 1][x]:
                count = count + 1
            if x > 0:
                if self.grid[y - 1][x - 1]:
                    count = count + 1
            if self.width > (x + 1):
                if self.grid[y - 1][x + 1]:
                    count = count + 1

        if x > 0:
            if self.grid[y][x - 1]:
                count = count + 1
        if self.width > (x + 1):
            if self.grid[y][x + 1]:
                count = count + 1

        if self.height > (y + 1):
            if self.grid[y + 1][x]:
                count = count + 1
            if x > 0:
                if self.grid[y + 1][x - 1]:
                    count = count + 1
            if self.width > (x + 1):
                if self.grid[y + 1][x + 1]:
                    count = count + 1

        return count

    def animate(self):
        """Game of Life turn"""
        nt = copy.deepcopy(self.grid)
        for y in range(0, self.height):
            for x in range(0, self.width):
                neighbours = self._live_neighbours(y, x)
                if self.grid[y][x] == 0:
                    if neighbours == 3:
                        nt[y][x] = 1
                else:
                    if (neighbours < 2) or (neighbours > 3):
                        nt[y][x] = 0
        if nt == self.grid and self.entropy <= 0:
            self.stop()
        self.grid = nt
        self.randomize(self.entropy)
        if self.is_empty:
            self.stop()


# Langtons Ant base
class _Ant:
    def __init__(self, x, y, direction, height=8, width=32):
        self.x = x
        self.y = y
        assert direction[0] in ["r", "l", "u", "d"]
        self.direction = direction[0]
        self.grid_height = height
        self.grid_width = width
        self.dead = False

    def move_forward(self):
        if self.direction == "r":
            self.x += 1
            if self.x == self.grid_width:
                self.dead = True
        elif self.direction == "d":
            self.y += 1
            if self.y == self.grid_height:
                self.dead = True
        elif self.direction == "l":
            self.x -= 1
            if self.x == -1:
                self.dead = True
        elif self.direction == "u":
            self.y -= 1
            if self.y == -1:
                self.dead = True

    def turn_right(self):
        if self.direction == "r":
            self.direction = "d"
        elif self.direction == "d":
            self.direction = "l"
        elif self.direction == "l":
            self.direction = "u"
        elif self.direction == "u":
            self.direction = "r"

    def turn_left(self):
        if self.direction == "r":
            self.direction = "u"
        elif self.direction == "d":
            self.direction = "r"
        elif self.direction == "l":
            self.direction = "d"
        elif self.direction == "u":
            self.direction = "l"


class _InfiniteAnt(_Ant):
    def move_forward(self):
        if self.direction == "r":
            self.x += 1
            if self.x == self.grid_width:
                self.x = 0
        elif self.direction == "d":
            self.y += 1
            if self.y == self.grid_height:
                self.y = 0
        elif self.direction == "l":
            self.x -= 1
            if self.x == -1:
                self.x = self.grid_width - 1
        elif self.direction == "u":
            self.y -= 1
            if self.y == -1:
                self.y = self.grid_height - 1


class _ReverseAnt(_Ant):
    def turn_left(self):
        super().turn_right()

    def turn_right(self):
        super().turn_left()


class _ReverseInfiniteAnt(_InfiniteAnt):
    def turn_left(self):
        super().turn_right()

    def turn_right(self):
        super().turn_left()


class LangtonsAnt(FacePlateAnimation):
    def __init__(self, ants=1, continuous=True, gen_reverse=False,
                 grid=None, bus=None):
        super().__init__(grid=grid, bus=bus)
        self.ants = []
        # if continuous loops around the board
        #   height + 1 -> 0
        #   width + 1 -> 0
        # else ant is removed
        self.continuous = continuous
        if isinstance(ants, int):
            # spawn N ants
            assert 0 <= ants < 256
            for i in range(ants):
                x = random.randint(0, self.width - 1)
                y = random.randint(0, self.height - 1)
                direction = random.choice(["u", "d", "l", "r"])
                reverse = False
                if gen_reverse:
                    reverse = random.choice([True, False])
                ant = self.ant_factory(x, y, direction, reverse)
                self.ants.append(ant)
        elif isinstance(ants, list):
            # Ant objects
            self.ants = ants
            for ant in self.ants:
                assert isinstance(ant, _Ant)
        else:
            raise ValueError

    def ant_factory(self, x, y, direction, reverse=False):
        # reverse ants exit black squares to the opposite direction
        if self.continuous:
            # loops around the board instead of dying
            if reverse:
                return _ReverseInfiniteAnt(x, y, direction)
            return _InfiniteAnt(x, y, direction)
        if reverse:
            return _ReverseAnt(x, y, direction)
        return _Ant(x, y, direction)

    def move_ants(self):
        # copy grid, multiple ants might want to flip same square
        # end result is the same so this is not a problem as long as it does
        # not change during iteration
        old_grid = copy.deepcopy(self.grid)

        for idx, ant in enumerate(self.ants):
            if self.ants[idx].dead:
                continue
            if old_grid[ant.y][ant.x] == 1:
                # black
                self.ants[idx].turn_left()
            else:
                # white
                self.ants[idx].turn_right()
            # flip color
            self.grid[ant.y][ant.x] = not old_grid[ant.y][ant.x]
            # update ant position
            self.ants[idx].move_forward()

    def animate(self):
        self.move_ants()
        # Stop condition -> all ants moved out of the board
        dead_ants = [ant for ant in self.ants if ant.dead]
        if len(dead_ants) == len(self.ants):
            self.stop()


# Game of Life Animations
class SpaceInvader(GoL):
    # This basically half "pulsar"
    str_grid = """
XXXXXXXXXXXX  XXX  XXXXXXXXXXXXX
XXXXXXXXXX X X X X X XXXXXXXXXXX
XXXXXXXX   XX  X  XX   XXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXX  XXX  XXXXXXXXXXXXX
XXXXXXXXXXXX XXXXX XXXXXXXXXXXXX
XXXXXXXXXXXX XXXXX XXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
"""


# Langton's Ant animations

# Single Ant
class LangtonsLineDisplacer(LangtonsAnt):
    # see pattern here
    # https://youtu.be/w6XQQhCgq5c?t=84

    def __init__(self, x=None, y=None, continuous=True, bus=None):
        super().__init__(0, continuous, bus=bus)
        x = x if x is not None else random.randint(0, self.width - 1)
        y = y if y is not None else random.randint(0, self.height - 1)
        ant = self.ant_factory(x, y - 1, "u")
        self.ants.append(ant)
        # create initial line
        for i in range(0, self.width):
            self.grid[y][i] = 1


# 2 Ants
class LangtonsAntsOscillator(LangtonsAnt):
    # see pattern here
    # https://youtu.be/w6XQQhCgq5c?t=103

    def __init__(self, x=None, y=None, bus=None):
        super().__init__(0, bus=bus)
        x1 = x2 = x if x is not None else self.width // 2
        y1 = y if y is not None else self.height // 2
        y2 = y1 - 1
        dir1 = "d"
        dir2 = "u"
        ant1 = self.ant_factory(x1, y1, dir1)
        ant2 = self.ant_factory(x2, y2, dir2)
        self.ants += [ant1, ant2]


class LangtonsAntsOscillator2(LangtonsAnt):
    def __init__(self, x=None, y=None, bus=None):
        super().__init__(0, bus=bus)
        x1 = x2 = x if x is not None else self.width // 2
        y1 = y if y is not None else self.height // 2
        y2 = y1 - 1
        dir1 = dir2 = "u"
        ant1 = self.ant_factory(x1, y1, dir1)
        ant2 = self.ant_factory(x2, y2, dir2)
        self.ants += [ant1, ant2]


class LangtonsAntsOscillator3(LangtonsAnt):
    def __init__(self, x=None, y=None, bus=None):
        super().__init__(0, bus=bus)
        x1 = x2 = x if x is not None else self.width // 2
        y1 = y if y is not None else self.height // 2
        y2 = y1 - 1
        dir1 = "l"
        dir2 = "r"
        ant1 = self.ant_factory(x1, y1, dir1)
        ant2 = self.ant_factory(x2, y2, dir2)
        self.ants += [ant1, ant2]


class LangtonsAntsOscillator4(LangtonsAnt):
    def __init__(self, x=None, y=None, bus=None):
        super().__init__(0, bus=bus)
        x1 = x2 = x if x is not None else self.width // 2
        y1 = y if y is not None else self.height // 2
        y2 = y1 - 1
        dir1 = dir2 = "l"
        ant1 = self.ant_factory(x1, y1, dir1)
        ant2 = self.ant_factory(x2, y2, dir2)
        self.ants += [ant1, ant2]


class LangtonsAntsOscillator5(LangtonsAnt):
    def __init__(self, x=None, y=None, bus=None):
        super().__init__(0, bus=bus)
        x1 = x2 = x if x is not None else self.width // 2
        y1 = y if y is not None else self.height // 2
        y2 = y1 - 1
        dir1 = "u"
        dir2 = "d"
        ant1 = self.ant_factory(x1, y1, dir1)
        ant2 = self.ant_factory(x2, y2, dir2)
        self.ants += [ant1, ant2]


class LangtonsAntTrail(LangtonsAnt):
    # see pattern here
    # https://youtu.be/w6XQQhCgq5c?t=159

    def __init__(self, x=None, y=None, bus=None):
        super().__init__(0, bus=bus)
        x = x if x is not None else random.randint(0, self.width - 1)
        y = y if y is not None else random.randint(0, self.height - 1)
        dir1 = "u"
        dir2 = "d"
        ant1 = self.ant_factory(x, y - 1, dir1)
        ant2 = self.ant_factory(x, y, dir2, reverse=True)
        self.ants += [ant1, ant2]


class LangtonsAntDotTransporter(LangtonsAnt):
    # https://youtu.be/w6XQQhCgq5c?t=171

    def __init__(self, x=None, y=None,  bus=None):
        super().__init__(0, bus=bus)
        x = x if x is not None else random.randint(0, self.width - 1)
        y = y if y is not None else random.randint(0, self.height - 1)
        dir1 = dir2 = "u"
        ant1 = self.ant_factory(x, y, dir1)
        ant2 = self.ant_factory(x + 1, y, dir2, reverse=True)
        self.ants += [ant1, ant2]
        # initial grid
        if y + 1 == self.height:
            self.grid[0][x + 1] = 1 # bellow anti-ant
        else:
            self.grid[y + 1][x + 1] = 1  # bellow anti-ant
        self.grid[y - 1][x] = 1  # above ant


# 1D / elementar automata

class ElementarAutomata(FacePlateAnimation):
    def __init__(self, direction="u", idx=0, seed=None, grid=None, bus=None):
        super().__init__(grid, bus)
        assert direction[0] in ["u", "d", "l", "r"]
        self.direction = direction[0]
        self.row = idx
        self.initial_state(seed)

    def initial_state(self, seed=None):
        if seed is not None:
            line = seed
        else:
            line = [0 for i in range(self.width)]
        self.grid[self.row] = line

    def rule(self):
        # process the line
        raise NotImplementedError

    def animate(self):
        old = copy.deepcopy(self.grid)
        new_line = self.rule()
        if self.direction == "u":
            self.move_up()
        elif self.direction == "d":
            self.move_down()
        elif self.direction == "l":
            self.move_left()
        elif self.direction == "r":
            self.move_right()
        self.grid[self.row] = new_line
        if old == self.grid:
            self.stop()


class SierpinskiTriangle(ElementarAutomata):
    def __init__(self, direction="u", seed=None, bus=None):
        assert direction[0] in ["u", "d"]
        if direction[0] == "u":
            idx = -1
        else:
            idx = 0
        super().__init__(direction, idx, seed=seed, bus=bus)

    def initial_state(self, seed=None):
        if seed is not None:
            line = seed
        else:
            line = [0 for i in range(self.width)]
            idx = self.width // 2
            line[idx] = 1
        self.grid[self.row] = line

    def rule(self):
        new_line = copy.deepcopy(self.grid[self.row])
        # handle middle
        for i in range(1, self.width - 1):
            left = self.grid[self.row][i - 1]
            right = self.grid[self.row][i + 1]
            if left and right:
                new_line[i] = 0
            elif left or right:
                new_line[i] = 1
            else:
                new_line[i] = 0

        # handle edges
        #if self.grid[0][1] == 1:
        #    new_line[0] = 1
        #else:
        #    new_line[0] = 0
        #if self.grid[0][-2] == 1:
        #    new_line[-1] = 1
        #else:
        #    new_line[-1] = 0
        return new_line


class Rule110(ElementarAutomata):
    def __init__(self, direction="u", seed=None, bus=None):
        assert direction[0] in ["u", "d"]
        if direction[0] == "u":
            idx = -1
        else:
            idx = 0
        super().__init__(direction, idx, seed=seed, bus=bus)

    def initial_state(self, seed=None):
        if seed is not None:
            line = seed
        else:
            line = [0 for i in range(self.width)]
            idx = self.width - 1
            line[idx] = 1
        self.grid[self.row] = line

    def rule(self):
        new_line = copy.deepcopy(self.grid[self.row])
        # handle middle
        for i in range(1, self.width - 1):
            left = self.grid[self.row][i - 1]
            mid = self.grid[self.row][i]
            right = self.grid[self.row][i + 1]
            if str(left) + str(mid) + str(right) in \
                    ["110", "101", "011", "010", "001"]:
                new_line[i] = 1
            else:
                new_line[i] = 0

        # handle edges
        if self.grid[0][1] == 1:
            new_line[0] = 1
        else:
            new_line[0] = 0
        if self.grid[0][-2] == 1:
            new_line[-1] = 1
        else:
            new_line[-1] = 0
        return new_line


if __name__ == "__main__":
    from time import sleep
    from ovos_utils.messagebus import get_mycroft_bus
    from time import sleep

    bus = get_mycroft_bus("192.168.1.70")

    a = Rule110(bus=bus)
    a.print()

    for grid in a:
        grid.print()
        grid.display(invert=False)
        sleep(0.5)


import subprocess


def is_mycroft_sj201():
    cmd = 'i2cdetect -y -a 1 0x04 0x04 | egrep "(04|UU)" | awk \'{print $2}\''
    out = subprocess.check_output(cmd, shell=True).strip()
    if out == b"04" or out == b"UU":
        return True
    return False


def is_respeaker_2mic():
    cmd = 'i2cdetect -y -a 1 0x1a 0x1a | egrep "(1a|UU)" | awk \'{print $2}\''
    out = subprocess.check_output(cmd, shell=True).strip()
    if out == b"1a" or out == b"UU":
        return True
    return False


def is_respeaker_4mic():
    cmd = 'i2cdetect -y -a 0x35 0x35 | egrep "(35|UU)" | awk \'{print $2}\''
    out = subprocess.check_output(cmd, shell=True).strip()
    if out == b"35" or out == b"UU":
        return True
    return False


def is_respeaker_6mic():
    cmd = 'i2cdetect -y -a 0x3b 0x3b | egrep "(3b|UU)" | awk \'{print $2}\''
    out = subprocess.check_output(cmd, shell=True).strip()
    if out == b"3b" or out == b"UU":
        return True
    return False


def is_adafruit():
    cmd = 'i2cdetect -y -a 0x4b 0x4b | egrep "(4b|UU)" | awk \'{print $2}\''
    out = subprocess.check_output(cmd, shell=True).strip()
    if out == b"4b" or out == b"UU":
        return True
    return False


def is_texas_tas5806():
    cmd = 'i2cdetect -y -a 0x2f 0x2f | egrep "(2f|UU)" | awk \'{print $2}\''
    out = subprocess.check_output(cmd, shell=True).strip()
    if out == b"2f" or out == b"UU":
        return True
    return False



from ovos_utils.lang.detect import detect_lang
import requests


def translate_text(text, lang="en-us", source_lang=None,
             url="https://libretranslate.com/translate"):
    """host it yourself https://github.com/uav4geo/LibreTranslate"""
    lang = lang.split("-")[0]
    if source_lang:
        source_lang = source_lang.split("-")[0]
    else:
        source_lang = detect_lang(text)
    r = requests.post(url, params={"q": text,
                                   "source": source_lang,
                                   "target": lang}).json()
    if r.get("error"):
        return None
    return r["translatedText"]



import platform
import pexpect
from socket import gethostname
from os.path import exists, join
from os import makedirs
import os
import random
import string
from ovos_utils.log import LOG

try:
    # pycryptodomex
    from Cryptodome.Cipher import AES
except ImportError:
    # pycrypto + pycryptodome
    try:
        from Crypto.Cipher import AES
    except:
        AES = None
try:
    from OpenSSL import crypto
except ImportError:
    crypto = None


def create_self_signed_cert(cert_dir, name="jarbas"):
    """
    If name.crt and name.key don't exist in cert_dir, create a new
    self-signed cert and key pair and write them into that directory.
    """
    if crypto is None:
        LOG.error("run pip install pyopenssl")
        raise ImportError
    CERT_FILE = name + ".crt"
    KEY_FILE = name + ".key"
    cert_path = join(cert_dir, CERT_FILE)
    key_path = join(cert_dir, KEY_FILE)

    if not exists(join(cert_dir, CERT_FILE)) \
            or not exists(join(cert_dir, KEY_FILE)):
        # create a key pair
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 1024)

        # create a self-signed cert
        cert = crypto.X509()
        cert.get_subject().C = "PT"
        cert.get_subject().ST = "Europe"
        cert.get_subject().L = "Mountains"
        cert.get_subject().O = "Jarbas AI"
        cert.get_subject().OU = "Powered by Mycroft-Core"
        cert.get_subject().CN = gethostname()
        cert.set_serial_number(random.randint(0, 2000))
        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(10 * 365 * 24 * 60 * 60)
        cert.set_issuer(cert.get_subject())
        cert.set_pubkey(k)
        # TODO don't use sha1
        cert.sign(k, 'sha1')
        if not exists(cert_dir):
            makedirs(cert_dir)
        open(cert_path, "wt").write(
            crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
        open(join(cert_dir, KEY_FILE), "wt").write(
            crypto.dump_privatekey(crypto.FILETYPE_PEM, k))

    return cert_path, key_path


def random_key(key_lenght=16):
    """Generate a random string of letters and digits """
    valid_chars = string.ascii_letters + string.digits
    return ''.join(random.choice(valid_chars) for i in range(key_lenght))


def encrypt(key, text, nonce=None):
    if AES is None:
        LOG.error("run pip install pycryptodomex")
        raise ImportError
    if not isinstance(text, bytes):
        text = bytes(text, encoding="utf-8")
    if not isinstance(key, bytes):
        key = bytes(key, encoding="utf-8")
    cipher = AES.new(key, AES.MODE_GCM, nonce=nonce)
    ciphertext, tag = cipher.encrypt_and_digest(text)
    return ciphertext, tag, cipher.nonce


def decrypt(key, ciphertext, tag, nonce):
    if AES is None:
        LOG.error("run pip install pycryptodomex")
        raise ImportError
    if not isinstance(key, bytes):
        key = bytes(key, encoding="utf-8")
    cipher = AES.new(key, AES.MODE_GCM, nonce)
    try:
        data = cipher.decrypt_and_verify(ciphertext, tag)
        text = data.decode(encoding="utf-8")
        return text
    except Exception as e:
        LOG.error("decryption failed, invalid key?")
        raise


def sudo_exec(cmdline, passwd="root"):
    osname = platform.system()
    if osname == 'Linux':
        prompt = r'\[sudo\] password for %s: ' % os.environ['USER']
    elif osname == 'Darwin':
        prompt = 'Password:'
    else:
        raise SystemError("Unsupported platform")

    child = pexpect.spawn(cmdline)
    idx = child.expect([prompt, pexpect.EOF], 3)
    if idx == 0:  # if prompted for the sudo password
        LOG.debug('sudo password was asked.')
        child.sendline(passwd)
        child.expect(pexpect.EOF)
    return child.before



from importlib.util import find_spec
from os.path import isfile, join, isdir

from json_database import JsonStorage

from ovos_utils.json_helper import load_commented_json, merge_dict
from ovos_utils.log import LOG
from ovos_utils.system import search_mycroft_core_location
from ovos_utils.xdg_utils import (
    xdg_config_home,
    xdg_config_dirs,
    xdg_data_home,
    xdg_data_dirs,
    xdg_cache_home
)


def get_xdg_config_dirs(folder=None):
    folder = folder or get_xdg_base()
    return [join(path, folder) for path in xdg_config_dirs() if isdir(join(path, folder))]


def get_xdg_data_dirs(folder=None):
    folder = folder or get_xdg_base()
    return [join(path, folder) for path in xdg_data_dirs() if isdir(join(path, folder))]


def get_xdg_config_save_path(folder=None):
    folder = folder or get_xdg_base()
    return join(xdg_config_home(), folder)


def get_xdg_data_save_path(folder=None):
    folder = folder or get_xdg_base()
    return join(xdg_data_home(), folder)


def get_xdg_cache_save_path(folder=None):
    folder = folder or get_xdg_base()
    return join(xdg_cache_home(), folder)


def get_ovos_config():
    config = {"xdg": True,
              "base_folder": "mycroft",
              "config_filename": "mycroft.conf"}

    try:
        config["default_config_path"] = find_default_config()
    except FileNotFoundError:
        pass  # not a mycroft device

    try:
        if isfile("/etc/OpenVoiceOS/ovos.conf"):
            config = merge_dict(config,
                                load_commented_json(
                                    "/etc/OpenVoiceOS/ovos.conf"))
        elif isfile("/etc/mycroft/ovos.conf"):
            config = merge_dict(config,
                                load_commented_json("/etc/mycroft/ovos.conf"))
    except:
        # tolerate bad json TODO proper exception (?)
        pass

    # This includes both the user config and
    # /etc/xdg/OpenVoiceOS/ovos.conf
    for p in get_xdg_config_dirs("OpenVoiceOS"):
        if isfile(join(p, "ovos.conf")):
            try:
                xdg_cfg = load_commented_json(join(p, "ovos.conf"))
                config = merge_dict(config, xdg_cfg)
            except:
                # tolerate bad json TODO proper exception (?)
                pass

    # let's check for derivatives specific configs
    # the assumption is that these cores are exclusive to each other,
    # this will never find more than one override
    # TODO this works if using dedicated .venvs what about system installs?
    cores = config.get("module_overrides") or {}
    for k in cores:
        if find_spec(k):
            config = merge_dict(config, cores[k])
            break
    else:
        subcores = config.get("submodule_mappings") or {}
        for k in subcores:
            if find_spec(k):
                config = merge_dict(config, cores[subcores[k]])
                break

    return config


def is_using_xdg():
    return get_ovos_config().get("xdg", True)


def get_xdg_base():
    return get_ovos_config().get("base_folder") or "mycroft"


def save_ovos_core_config(new_config):
    OVOS_CONFIG = join(get_xdg_config_save_path("OpenVoiceOS"),
                       "ovos.conf")
    cfg = JsonStorage(OVOS_CONFIG)
    cfg.update(new_config)
    cfg.store()
    return cfg


def set_xdg_base(folder_name):
    LOG.info(f"XDG base folder set to: '{folder_name}'")
    save_ovos_core_config({"base_folder": folder_name})


def set_config_filename(file_name, core_folder=None):
    if core_folder:
        set_xdg_base(core_folder)
    LOG.info(f"config filename set to: '{file_name}'")
    save_ovos_core_config({"config_filename": file_name})


def set_default_config(file_path=None):
    file_path = file_path or find_default_config()
    LOG.info(f"default config file changed to: {file_path}")
    save_ovos_core_config({"default_config_path": file_path})


def find_default_config():
    mycroft_root = search_mycroft_core_location()
    if not mycroft_root:
        raise FileNotFoundError("Couldn't find mycroft core root folder.")
    return join(mycroft_root, "mycroft", "configuration", "mycroft.conf")


def find_user_config():
    if is_using_xdg():
        path = join(get_xdg_config_save_path(), get_config_filename())
        if isfile(path):
            return path
    old, path = get_config_locations(default=False, web_cache=False,
                                     system=False, old_user=True,
                                     user=True)
    if isfile(path):
        return path
    if isfile(old):
        return old
    # mark1 runs as a different user
    sysconfig = MycroftSystemConfig()
    platform_str = sysconfig.get("enclosure", {}).get("platform", "")
    if platform_str == "mycroft_mark_1":
        path = "/home/mycroft/.mycroft/mycroft.conf"
    return path


def get_config_locations(default=True, web_cache=True, system=True,
                         old_user=True, user=True):
    locs = []
    ovos_cfg = get_ovos_config()
    if default:
        locs.append(ovos_cfg["default_config_path"])
    if system:
        locs.append(f"/etc/{ovos_cfg['base_folder']}/{ovos_cfg['config_filename']}")
    if web_cache:
        locs.append(get_webcache_location())
    if old_user:
        locs.append(f"~/.{ovos_cfg['base_folder']}/{ovos_cfg['config_filename']}")
    if user:
        if is_using_xdg():
            locs.append(f"{get_xdg_config_save_path()}/{ovos_cfg['config_filename']}")
        else:
            locs.append(f"~/.{ovos_cfg['base_folder']}/{ovos_cfg['config_filename']}")
    return locs


def get_webcache_location():
    return join(get_xdg_config_save_path(), 'web_cache.json')


def get_xdg_config_locations():
    # This includes both the user config and
    # /etc/xdg/mycroft/mycroft.conf
    xdg_paths = list(reversed(
        [join(p, get_config_filename())
         for p in get_xdg_config_dirs()]
    ))
    return xdg_paths


def get_config_filename():
    return get_ovos_config().get("config_filename") or "mycroft.conf"


def set_config_name(name, core_folder=None):
    # TODO deprecate, was only out in a couple versions
    # renamed to match HolmesV
    set_config_filename(name, core_folder)


def read_mycroft_config():
    conf = LocalConf("tmp/dummy.conf")
    conf.merge(MycroftDefaultConfig())
    conf.merge(MycroftSystemConfig())
    conf.merge(MycroftUserConfig())
    return conf


def update_mycroft_config(config, path=None):
    if path is None:
        conf = MycroftUserConfig()
    else:
        conf = LocalConf(path)
    conf.merge(config)
    conf.store()
    return conf


class LocalConf(JsonStorage):
    """
        Config dict from file.
    """
    allow_overwrite = True

    def __init__(self, path=None):
        super(LocalConf, self).__init__(path)


class ReadOnlyConfig(LocalConf):
    """ read only  """

    def __init__(self, path, allow_overwrite=False):
        super().__init__(path)
        self.allow_overwrite = allow_overwrite

    def reload(self):
        old = self.allow_overwrite
        self.allow_overwrite = True
        super().reload()
        self.allow_overwrite = old

    def __setitem__(self, key, value):
        if not self.allow_overwrite:
            raise PermissionError
        super().__setitem__(key, value)

    def __setattr__(self, key, value):
        if not self.allow_overwrite:
            raise PermissionError
        super().__setattr__(key, value)

    def merge(self, *args, **kwargs):
        if not self.allow_overwrite:
            raise PermissionError
        super().merge(*args, **kwargs)

    def store(self, path=None):
        if not self.allow_overwrite:
            raise PermissionError
        super().store(path)


class MycroftUserConfig(LocalConf):
    def __init__(self):
        path = find_user_config()
        super().__init__(path)


class MycroftDefaultConfig(ReadOnlyConfig):
    def __init__(self):
        path = get_ovos_config()["default_config_path"]
        super().__init__(path)
        if not self.path or not isfile(self.path):
            LOG.debug(f"mycroft root path not found, could not load default .conf: {self.path}")

    def set_root_config_path(self, root_config):
        # in case we got it wrong / non standard
        self.path = root_config
        self.reload()


class MycroftSystemConfig(ReadOnlyConfig):
    def __init__(self, allow_overwrite=False):
        path = get_config_locations(default=False, web_cache=False,
                                    system=True, old_user=False,
                                    user=False)[0]
        super().__init__(path, allow_overwrite)


class MycroftXDGConfig(LocalConf):
    def __init__(self):
        path = get_config_locations(default=False, web_cache=False,
                                    system=False, old_user=False,
                                    user=True)[0]
        super().__init__(path)



from ovos_utils.enclosure.mark1 import Mark1EnclosureAPI
from ovos_utils.messagebus import get_mycroft_bus
from ovos_utils.colors import Color
from ovos_utils import rotate_list
from time import sleep


class EyePixel:
    def __init__(self, index, api, color=None):
        self.index = index
        self.api = api
        self.color = color or Color()

    @property
    def rgb(self):
        return self.color.rgb255

    def sync_color(self):
        r, g, b = self.api.get_eyes_pixel_color(self.index)
        color = Color.from_rgb(r, g, b)
        self.change_color(color)

    def update_color(self):
        self.change_color(self.color)

    def change_color(self, name):
        if isinstance(name, str):
            self.color = Color.from_name(name)
        elif isinstance(name, Color):
            self.color = name
        else:
            raise ValueError("not a Color object")
        r, g, b = self.rgb
        self.api.eyes_setpixel(self.index, r, g, b)

    def set_saturation(self, value):
        self.color.set_saturation(value)
        self.update_color()

    def set_luminance(self, value):
        self.color.set_luminance(value)
        self.update_color()

    def set_hue(self, value):
        self.color.set_hue(value)
        self.update_color()

    def __repr__(self):
        return "Pixel_" + str(self.index) + ":" + self.color.color_description


class Eye(list):
    def __init__(self, pixel_range, bus=None, color=None):
        super().__init__()
        self.bus = bus or get_mycroft_bus()
        self.api = Mark1EnclosureAPI(self.bus)
        for idx in range(pixel_range[0], pixel_range[1]):
            pixel = EyePixel(idx, self.api)
            self.append(pixel)
        self.color = Color()
        if color:
            self.change_color(color)
        else:
            self.sync_color()

    def sync_color(self):
        for p in self:
            p.sync_color()
            sleep(0.05)

    def update_color(self):
        for p in self:
            p.update_color()
            sleep(0.05)

    def change_color(self, name):
        if isinstance(name, str):
            self.color = Color.from_name(name)
        elif isinstance(name, Color):
            self.color = name
        else:
            raise ValueError("not a Color object")
        for led in self:
            led.change_color(self.color)
            # writer bugs out if messages sent too fast
            sleep(0.05)

    def saturation_spin(self, speed=0.05):
        values = []
        for idx, pixel in enumerate(self):
            sat = 0.09 * idx
            pixel.set_saturation(sat)
            values.append(sat)
            sleep(speed)

        while True:
            values = rotate_list(values, -1)
            for idx, value in enumerate(values):
                self[idx].set_saturation(value)
                sleep(speed)

    def luminance_spin(self, speed=0.05):
        values = []
        for idx, pixel in enumerate(self):
            sat = 0.05 * idx
            pixel.set_luminance(sat)
            values.append(sat)
            sleep(speed)

        while True:
            values = rotate_list(values, -1)
            for idx, value in enumerate(values):
                self[idx].set_luminance(value)
                sleep(speed)

    def hue_spin(self, speed=0.05):
        values = []
        for idx, pixel in enumerate(self):
            sat = 0.083 * idx
            pixel.set_hue(sat)
            values.append(sat)
            sleep(speed)

        while True:
            values = rotate_list(values, -1)
            for idx, value in enumerate(values):
                self[idx].set_hue(value)
                sleep(speed)

    def set_hue(self, hue):
        for pixel in self:
            pixel.color.set_hue(hue)
            pixel.update_color()

    def set_luminance(self, value):
        for pixel in self:
            pixel.color.set_luminance(value)
        self.update_color()

    def set_saturation(self, value):
        for pixel in self:
            pixel.color.set_saturation(value)
        self.update_color()

    def on(self):
        self.set_luminance(1)

    def off(self):
        self.set_luminance(0)

    def blink_once(self):
        """
        Make the eye blink
        """
        raise NotImplementedError

    def blink(self, speed=0.5):
        """
        Make the right eye blink in a loop
        """
        while True:
            self.blink_once()
            sleep(speed)


class RightEye(Eye):
    def __init__(self, bus, color=None):
        super().__init__(bus=bus, pixel_range=(12, 24), color=color)

    def sync_color(self):
        pixels = self.api.get_eyes_color()[12:]
        for idx, (r, g, b) in enumerate(pixels):
            self[idx].color = Color.from_rgb(r, g, b)
        self.update_color()

    def blink_once(self):
        """
        Make the right eye blink
        """
        self.api.eyes_blink("r")


class LeftEye(Eye):
    def __init__(self, bus, color=None):
        super().__init__(bus=bus, pixel_range=(0, 12), color=color)

    def sync_color(self):
        pixels = self.api.get_eyes_color()[:12]
        for idx, (r, g, b) in enumerate(pixels):
            self[idx].color = Color.from_rgb(r, g, b)
        self.update_color()

    def blink_once(self):
        """
        Make the left eye blink
        """
        self.api.eyes_blink("l")


class Eyes(list):
    def __init__(self, bus=None, color=None):
        super().__init__()
        self.bus = bus or get_mycroft_bus()
        self.api = Mark1EnclosureAPI(self.bus)
        self.right = RightEye(self.bus)
        self.left = LeftEye(self.bus)
        self.color = Color()
        if color:
            self.change_color(color)
        else:
            self.sync_color()

    def __getitem__(self, item):
        assert isinstance(item, int)
        assert 0 <= item <= 23
        if item < 12:
            return self.left[item]
        return self.right[item - 12]

    def __setitem__(self, key, value):
        assert isinstance(key, int)
        assert 0 <= key <= 23
        if key < 12:
            self.left[key] = value
        self.right[key] = value

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        return len(self.left) + len(self.right)

    def sync_color(self):
        """ updates internal color value to current color """
        pixels = self.api.get_eyes_color()
        for idx, (r, g, b) in enumerate(pixels):
            self[idx].color = Color.from_rgb(r, g, b)

    def update_color(self):
        """ updates arduino color to current pixels """
        for i in range(len(self) // 2):
            self.left[i].update_color()
            sleep(0.05)
            self.right[i].update_color()
            sleep(0.05)

    def change_color(self, name):
        """ changes color of both eyes """
        if isinstance(name, str):
            self.color = Color.from_name(name)
        elif isinstance(name, Color):
            self.color = name
        else:
            raise ValueError("not a Color object")
        r, g, b = self.color.rgb255
        self.api.eyes_color(r, g, b)
        for idx in range(len(self)):
            self[idx].color = self.color

    # animations
    def saturation_spin(self, speed=0.05):
        values = []
        for idx in range(len(self) // 2):
            sat = 0.09 * idx
            values.append(sat)
            self.left[idx].set_saturation(sat)
            sleep(0.03)
            self.right[idx].set_saturation(sat)

        while True:
            values = rotate_list(values, -1)
            for idx, value in enumerate(values):
                self.left[idx].set_saturation(value)
                sleep(0.03)
                self.right[idx].set_saturation(value)
                sleep(speed)

    def luminance_spin(self, speed=0.05):
        values = []
        for idx in range(len(self) // 2):
            sat = 0.05 * idx
            values.append(sat)
            self.left[idx].set_luminance(sat)
            sleep(0.03)
            self.right[idx].set_luminance(sat)

        while True:
            values = rotate_list(values, -1)
            for idx, value in enumerate(values):
                self.left[idx].set_luminance(value)
                sleep(0.03)
                self.right[idx].set_luminance(value)
                sleep(speed)

    def hue_spin(self, speed=0.05):
        values = []
        for idx in range(len(self) // 2):
            sat = 0.083 * idx
            values.append(sat)
            self.left[idx].set_hue(sat)
            sleep(0.03)
            self.right[idx].set_hue(sat)

        while True:
            values = rotate_list(values, -1)
            for idx, value in enumerate(values):
                for pixel in self:
                    print(pixel)
                self.left[idx].set_hue(value)
                sleep(0.03)
                self.right[idx].set_hue(value)
                sleep(speed)

    def flash(self, speed=0.2):
        while True:
            sleep(speed)
            self.on()
            sleep(speed)
            self.off()

    def rainbow_flash(self, speed=0.2):
        colors = ["red", "orange", "yellow", "green", "cyan", "blue",
                  "violet", "purple"]
        while True:
            for color in colors:
                sleep(speed)
                self.off()
                self.change_color(color)
                sleep(speed)
                self.on()

    def beacon(self, speed=0.1):
        values = [i + i for i in range(30)]
        while True:
            for value in values:
                self.set_brightness(value)
                sleep(speed)
            values.reverse()

    def rainbow_beacon(self, speed=0.1):
        values = [i + i for i in range(30)]
        values += reversed(values)
        colors = ["red", "orange", "yellow", "green", "cyan", "blue",
                  "violet", "purple"]
        self.set_brightness(0)
        while True:
            for color in colors:
                for value in values:
                    sleep(speed)
                    self.set_brightness(value)
                self.change_color(color)

    # Arduino API
    def set_hue(self, hue):
        self.right.set_hue(hue)
        sleep(0.05)
        self.left.set_hue(hue)

    def set_brightness(self, level):
        """
        Set the brightness of the eyes in the display.
        Args:
            level (int): 1-30, bigger numbers being brighter
        """
        self.api.eyes_brightness(level)

    def spin(self):
        self.api.eyes_spin()

    def timed_spin(self, length):
        self.api.eyes_timed_spin(length)

    def reset(self):
        self.api.eyes_reset()

    def fill_once(self, percent):
        """
        Use the eyes as a type of progress meter
        Args:
            percent (int): 0-49 fills the right eye, 50-100 also covers left
        """
        self.api.eyes_fill(percent)

    def look(self, side):
        """Make the eyes look to the given side
        Args:
            side (str): 'r' for right
                        'l' for left
                        'u' for up
                        'd' for down
                        'c' for crossed
        """
        self.api.eyes_look(side)

    def look_right(self):
        self.look("r")

    def look_left(self):
        self.look("l")

    def look_up(self):
        self.look("u")

    def look_down(self):
        self.look("d")

    def cross(self):
        self.look("c")

    def narrow(self):
        """Make the eyes look narrow, like a squint"""
        self.api.eyes_narrow()

    def on(self):
        """Illuminate or show the eyes."""
        self.api.eyes_on()

    def off(self):
        """Turn off or hide the eyes."""
        self.api.eyes_off()

    def blink_once(self, side="b"):
        """Make the eyes blink
        Args:
            side (str): 'r', 'l', or 'b' for 'right', 'left' or 'both'
        """
        self.api.eyes_blink(side)

    def blink_right_once(self):
        self.right.blink_once()

    def blink_left_once(self):
        self.left.blink_once()

    def blink(self, speed=0.5):
        """
        Make the eyes blink in a loop
        """
        while True:
            self.blink_once()
            sleep(speed)

    def blink_right(self, speed=0.5):
        """
        Make the right eye blink in a loop
        """
        self.right.blink(speed)

    def blink_left(self, speed=0.5):
        """
        Make the left eyes blink in a loop
        """
        self.left.blink(speed)

    def blink_alternate(self, speed=0.5):
        """
        Make the eyes blink in a loop
        """
        while True:
            self.blink_right_once()
            sleep(speed)
            self.blink_left_once()
            sleep(speed)

    def up_down(self, speed=0.8):
        """
        Make the eyes blink in a loop
        """
        while True:
            self.look_up()
            sleep(speed)
            self.look_down()
            sleep(speed)

    def left_right(self, speed=0.8):
        """
        Make the eyes blink in a loop
        """
        while True:
            self.look_left()
            sleep(speed)
            self.look_right()
            sleep(speed)

    def fill(self, speed=0.1):
        values = [i for i in range(101)]
        values += reversed(values)
        while True:
            for percent in values:
                self.fill_once(percent)
                sleep(speed)

    def rainbow_fill(self, speed=0.1):
        values = [i for i in range(101)]
        values += reversed(values)
        colors = ["red", "orange", "yellow", "green", "cyan", "blue",
                  "violet", "purple"]
        while True:
            for color in colors:
                for percent in values:
                    self.fill_once(percent)
                    sleep(speed)
                    if percent == 100:
                        self.change_color(color)



if __name__ == "__main__":
    bus = get_mycroft_bus("192.168.1.70")
    eyes = Eyes(bus)
    eyes.hue_spin()



from colorsys import rgb_to_yiq, rgb_to_hls, yiq_to_rgb, hls_to_rgb, rgb_to_hsv, hsv_to_rgb
from colour import Color as _Color
from ovos_utils import camel_case_split


class UnrecognizedColorName(ValueError):
    """ No color defined with this name """


class Color(_Color):
    """ A well defined Color, just the way computers love colors"""
    @property
    def name(self):
        if self.web != self.hex:
            return camel_case_split(self.web).lower()
        return None

    @staticmethod
    def from_name(name):
        try:
            name = name.lower().replace(" ", "").strip()
            return Color(name)
        except:
            # try to parse color description
            color = ColorOutOfSpace()

            if "bright" in name:
                color.set_luminance(0.7)
            if "dark" in name:
                color.set_luminance(0.3)

            if "light" in name:
                color.set_saturation(0.4)
            if "grey" in name or "gray" in name:
                color.set_saturation(0.25)

            if "black" in name:
                color.set_luminance(0.1)
            if "white" in name:
                color.set_luminance(1)

            red = 0.0
            orange = 0.10
            yellow = 0.16
            green = 0.33
            cyan = 0.5
            blue = 0.66
            violet = 0.83

            if "orange" in name:
                color.hue = orange
            elif "yellow" in name:
                color.hue = yellow
            elif "green" in name:
                color.hue = green
            elif "cyan" in name:
                color.hue = cyan
            elif "blue" in name:
                color.hue = blue
            elif "violet" in name:
                color.hue = violet
            else:
                color.hue = red

            return color

    @property
    def main_color(self):
        """
        reduce to 1 color of the following:
        - grey
        - black
        - white
        - orange
        - yellow
        - green
        - cyan
        - blue
        - violet
        - red
        """
        if self.saturation <= 0.3:
            return Color("grey")
        if self.luminance <= 0.15:
            return Color("black")
        elif self.luminance >= 0.85:
            return Color("white")

        thresh = 0.5
        orange = 0.10
        yellow = 0.16
        green = 0.33
        cyan = 0.5
        blue = 0.66
        violet = 0.83

        if orange - thresh <= self.hue <= orange + thresh:
            return Color("orange")
        elif yellow - thresh <= self.hue <= yellow + thresh:
            return Color("yellow")
        elif green - thresh <= self.hue <= green + thresh:
            return Color("green")
        elif cyan - thresh <= self.hue <= cyan + thresh:
            return Color("cyan")
        elif blue - thresh <= self.hue <= blue + thresh:
            return Color("blue")
        elif violet - thresh <= self.hue <= violet + thresh:
            return Color("violet")
        else:
            return Color("red")

    @property
    def color_description(self):
        if self.web != self.hex:
            return camel_case_split(self.web).lower()
        name = ""
        # light vs dark
        if self.luminance <= 0.3:
            name += "dark "
        elif self.luminance >= 0.6:
            name += "bright "

        # hue
        # R >== B >= G Red
        if self.red >= self.blue >= self.green:
            name += "red-ish "
        # R >== G >= = B 	Orange
        elif self.red >= self.green >= self.blue:
            name += "orange-ish "
        # G >= R >== B Yellow
        elif self.green >= self.red >= self.blue:
            name += "yellow-ish "
        # G >== B >= R  Green
        elif self.green >= self.blue >= self.red:
            name += "green-ish "
        # B >= G >= R  Blue
        elif self.blue >= self.green >= self.red:
            name += "blue-ish "
        # B >= R >== G  Violet
        elif self.blue >= self.red >= self.green:
            name += "purple-ish "

        # luminance
        if self.luminance <= 0.15:
            name += "black-ish "
        elif self.luminance >= 0.85:
            name += "white-ish "

        # saturation
        if self.saturation >= 0.85:
            name += "intense "
        name += self.main_color.name + " color"

        return name

    #### HEX ####
    @staticmethod
    def from_hex(hex_value):
        return Color(hex_value)

    #### RGB ####
    @property
    def rgb255(self):
        return (int(self.red * 255),
                int(self.green * 255),
                int(self.blue * 255))

    def rgb_percent(self):
        return self.rgb

    @staticmethod
    def from_rgb(r, g, b):
        return Color(rgb=(r / 255, g / 255, b / 255))

    @staticmethod
    def from_rgb_percent(r, g, b):
        if isinstance(r, str) or isinstance(g, str) or isinstance(b, str):
            r = float(r.replace("%", ""))
            g = float(g.replace("%", ""))
            b = float(b.replace("%", ""))
        return Color(rgb=(r, g, b))

    #### HSV ####
    @staticmethod
    def from_hsv(h, s, v):
        r, g, b = hsv_to_rgb(h, s, v)
        return Color.from_rgb(r, g, b)

    @property
    def hsv(self):
        return rgb_to_hsv(self.red, self.green, self.blue)

    #### HLS ####
    @staticmethod
    def from_hls(h, l, s):
        r, g, b = hls_to_rgb(h, l, s)
        return Color.from_rgb(r, g, b)

    @property
    def hls(self):
        return rgb_to_hls(self.red, self.green, self.blue)

    #### YIQ ####
    @staticmethod
    def from_yiq(y, i, q):
        r, g, b = yiq_to_rgb(y, i, q)
        return Color.from_rgb(r, g, b)

    @property
    def yiq(self):
        return rgb_to_yiq(self.red, self.green, self.blue)

    def __str__(self):
        return self.hex_l


class ColorOutOfSpace(Color):
    """ Some Human described this color, but humans suck at this"""
    @property
    def name(self):
        # H.P. Lovecraft - https://www.youtube.com/watch?v=4liRxrDzS5I
        # return "The Color Out of Space"
        return self.hex


if __name__ == "__main__":
    black = Color()
    assert black == Color("black")

    # NOTE this is the web name
    try:
        color = Color("dark green")
    except ValueError:
        color = Color("DarkGreen")
        color.from_name("dark green")  # use this one instead

    assert color.web == "DarkGreen"
    assert color.name == "dark green"

    white = Color.from_name("white")
    assert white == Color.from_rgb(255, 255, 255)

    color = Color.from_name("NNNNNNNNNNNNNNNNNN")
    assert isinstance(color, ColorOutOfSpace)
    assert color.name != "black"
    assert color.color_description == "black"

    color = Color.from_rgb(0, 120, 240)
    assert color.name is None
    color.set_saturation(0.3)
    color.set_luminance(0.7)
    assert color.color_description == "bright blue-ish gray color"
    assert color.rgb255 == (156, 179, 202)

    aprox_color = Color.from_name("bright blue-ish grey color")
    assert aprox_color.color_description == "bright blue-ish gray color"
    assert aprox_color.rgb255 != color.rgb255
    assert aprox_color.rgb255 == (160, 161, 198)



import os
import random
import re
from os.path import join
from pathlib import Path

from ovos_utils.bracket_expansion import expand_options
from ovos_utils.configuration import read_mycroft_config
from ovos_utils.file_utils import resolve_resource_file
from ovos_utils.lang import translate_word
from ovos_utils.log import LOG


class MustacheDialogRenderer:
    """A dialog template renderer based on the mustache templating language."""

    def __init__(self):
        self.templates = {}
        self.recent_phrases = []

        # TODO magic numbers are bad!
        self.max_recent_phrases = 3
        # We cycle through lines in .dialog files to keep Mycroft from
        # repeating the same phrase over and over. However, if a .dialog
        # file only contains a few entries, this can cause it to loop.
        #
        # This offset will override max_recent_phrases on very short .dialog
        # files. With the offset at 2, .dialog files with 3 or more lines will
        # be managed to avoid repetition, but .dialog files with only 1 or 2
        # lines will be unaffected. Dialog should never get stuck in a loop.
        self.loop_prevention_offset = 2

    def load_template_file(self, template_name, filename):
        """Load a template by file name into the templates cache.

        Args:
            template_name (str): a unique identifier for a group of templates
            filename (str): a fully qualified filename of a mustache template.
        """
        with open(filename, 'r', encoding='utf8') as f:
            for line in f:
                template_text = line.strip()
                # Skip all lines starting with '#' and all empty lines
                if (not template_text.startswith('#') and
                        template_text != ''):
                    if template_name not in self.templates:
                        self.templates[template_name] = []

                    # convert to standard python format string syntax. From
                    # double (or more) '{' followed by any number of
                    # whitespace followed by actual key followed by any number
                    # of whitespace followed by double (or more) '}'
                    template_text = re.sub(r'\{\{+\s*(.*?)\s*\}\}+', r'{\1}',
                                           template_text)

                    self.templates[template_name].append(template_text)

    def render(self, template_name, context=None, index=None):
        """
        Given a template name, pick a template and render it using the context.
        If no matching template exists use template_name as template.

        Tries not to let Mycroft say exactly the same thing twice in a row.

        Args:
            template_name (str): the name of a template group.
            context (dict): dictionary representing values to be rendered
            index (int): optional, the specific index in the collection of
                templates

        Returns:
            str: the rendered string
        """
        context = context or {}
        if template_name not in self.templates:
            # When not found, return the name itself as the dialog
            # This allows things like render("record.not.found") to either
            # find a translation file "record.not.found.dialog" or return
            # "record not found" literal.
            return template_name.replace('.', ' ')

        # Get the .dialog file's contents, minus any which have been spoken
        # recently.
        template_functions = self.templates.get(template_name)

        if index is None:
            template_functions = ([t for t in template_functions
                                   if t not in self.recent_phrases] or
                                  template_functions)
            line = random.choice(template_functions)
        else:
            line = template_functions[index % len(template_functions)]
        # Replace {key} in line with matching values from context
        line = line.format(**context)
        line = random.choice(expand_options(line))

        # Here's where we keep track of what we've said recently. Remember,
        # this is by line in the .dialog file, not by exact phrase
        self.recent_phrases.append(line)
        if (len(self.recent_phrases) >
                min(self.max_recent_phrases, len(self.templates.get(
                    template_name)) - self.loop_prevention_offset)):
            self.recent_phrases.pop(0)
        return line


def load_dialogs(dialog_dir, renderer=None):
    """Load all dialog files within the specified directory.

    Args:
        dialog_dir (str): directory that contains dialog files

    Returns:
        a loaded instance of a dialog renderer
    """
    if renderer is None:
        renderer = MustacheDialogRenderer()

    directory = Path(dialog_dir)
    if not directory.exists() or not directory.is_dir():
        LOG.warning('No dialog files found: {}'.format(dialog_dir))
        return renderer

    for path, _, files in os.walk(str(directory)):
        for f in files:
            if f.endswith('.dialog'):
                renderer.load_template_file(f.replace('.dialog', ''),
                                            join(path, f))
    return renderer


def get_dialog(phrase, lang=None, context=None):
    """Looks up a resource file for the given phrase.

    If no file is found, the requested phrase is returned as the string. This
    will use the default language for translations.

    Args:
        phrase (str): resource phrase to retrieve/translate
        lang (str): the language to use
        context (dict): values to be inserted into the string

    Returns:
        str: a randomized and/or translated version of the phrase
    """

    if not lang:
        try:
            conf = read_mycroft_config()
            lang = conf.get('lang')
        except FileNotFoundError:
            lang = "en-us"

    filename = join('text', lang.lower(), phrase + '.dialog')
    template = resolve_resource_file(filename)
    if not template:
        LOG.debug('Resource file not found: {}'.format(filename))
        return phrase

    stache = MustacheDialogRenderer()
    stache.load_template_file('template', template)
    if not context:
        context = {}
    return stache.render('template', context)


def join_list(items, connector, sep=None, lang=''):
    """ Join a list into a phrase using the given connector word
    Examples:
        join_list([1,2,3], "and") ->  "1, 2 and 3"
        join_list([1,2,3], "and", ";") ->  "1; 2 and 3"
    Args:
        items (array): items to be joined
        connector (str): connecting word (resource name), like "and" or "or"
        sep (str, optional): separator character, default = ","
        lang (str, optional): an optional BCP-47 language code, if omitted
                              the default language will be used.
    Returns:
        str: the connected list phrase
    """

    if not items:
        return ""
    if len(items) == 1:
        return str(items[0])

    if not sep:
        sep = ", "
    else:
        sep += " "
    return (sep.join(str(item) for item in items[:-1]) +
            " " + translate_word(connector, lang) +
            " " + items[-1])



import unittest
from copy import deepcopy
from ovos_utils.json_helper import merge_dict, is_compatible_dict,\
    nested_delete, nested_set, nested_get, flatten_dict, flattened_get, \
    invert_dict, flattened_set, flattened_delete


class TestJsonHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.base_dict = {"one": 1, "two": 2, "three": 3,
                          "four": ["foo", "bar", "baz"], "five": 50}
        self.delta_dict = {"two": 2, "three": 30,
                           "four": [4, 5, 6, "foo"], "five": None}

    def test_invert_dict(self):
        dct = {'a': 1, 'b': "d"}
        self.assertEqual(invert_dict(dct), {1: "a", 'd': "b"})

    def test_flatten_dict(self):
        dct = {'a': 1, 'b': {'c': 1, 'd': 2}}

        self.assertEqual(flatten_dict(dct),
                         {'a': 1, 'b:c': 1, 'b:d': 2})
        self.assertEqual(flatten_dict(dct, separator="."),
                         {'a': 1, 'b.c': 1, 'b.d': 2})

        self.assertEqual(flattened_get(dct, "a"), 1)
        self.assertEqual(flattened_get(dct, "b:c"), 1)
        self.assertEqual(flattened_get(dct, "b:d"), 2)
        self.assertEqual(flattened_get(dct, "b.d", separator="."), 2)

        # methods below modify original dict
        flattened_set(dct, "b:d", 5)
        self.assertEqual(dct, {'a': 1, 'b': {'c': 1, "d": 5}})
        flattened_set(dct, "b.c", 5, separator=".")
        self.assertEqual(dct, {'a': 1, 'b': {'c': 5, "d": 5}})

        flattened_delete(dct, "b:d")
        self.assertEqual(dct, {'a': 1, 'b': {'c': 5}})
        flattened_delete(dct, "b.c", separator=".")
        self.assertEqual(dct, {'a': 1, 'b': {}})

    def test_nested_get(self):
        dct = {'a': 1, 'b': {'c': 1, 'd': 2}}
        self.assertEqual(nested_get(dct, ["a"]), 1)
        self.assertEqual(nested_get(dct, ["b"]), {'c': 1, 'd': 2})
        self.assertEqual(nested_get(dct, ["b", "c"]), 1)
        self.assertEqual(nested_get(dct, ["b", "d"]), 2)

    def test_nested_set(self):
        dct = {'a': 1, 'b': {'c': 1, 'd': 2}}

        nested_set(dct, ["a"], 2)
        self.assertEqual(dct, {'a': 2, 'b': {'c': 1, 'd': 2}})
        nested_set(dct, ["b", "c"], 5)
        self.assertEqual(dct, {'a': 2, 'b': {'c': 5, 'd': 2}})
        nested_set(dct, ["b", "d"], 5)
        self.assertEqual(dct, {'a': 2, 'b': {'c': 5, 'd': 5}})

    def test_nested_delete(self):
        dct = {'a': 1, 'b': {'c': 1, 'd': 2}}

        nested_delete(dct, ["a"])
        self.assertEqual(dct, {'b': {'c': 1, 'd': 2}})
        nested_delete(dct, ["b", "c"])
        self.assertEqual(dct, {'b': {'d': 2}})
        nested_delete(dct, ["b", "d"])
        self.assertEqual(dct, {'b': {}})
        nested_delete(dct, ["b"])
        self.assertEqual(dct, {})

        dct = {'a': 1, 'b': {'c': 1, 'd': 2}}
        nested_delete(dct, ["b"])
        self.assertEqual(dct, {"a": 1})

    def test_compatible_dict(self):
        d0 = {"k2": 2, "k3": [0, 0], "k4": {"not": "yeah"}}
        d1 = {"k": "val", "k2": 1, "k3": [], "k5": None}
        d2 = {"k4": {"a": 0}}
        d3 = {"k2": "2", "k4": {"a": ["type_changed"]}}

        self.assertEqual(is_compatible_dict(d0, d1), True)
        self.assertEqual(is_compatible_dict(d0, d2), True)
        self.assertEqual(is_compatible_dict(d1, d2), True)
        self.assertEqual(is_compatible_dict(d0, d3), False) # k2 type changed
        self.assertEqual(is_compatible_dict(d1, d3), False) # k4 type changed
        self.assertEqual(is_compatible_dict(d2, d3), False) # k4 dicts incompatible

    def test_merge_dict(self):
        self.assertEqual(
            merge_dict(deepcopy(self.base_dict), deepcopy(self.delta_dict)),
            {"one": 1, "two": 2, "three": 30, "four": [
                4, 5, 6, "foo"], "five": None}
        )

    def test_merge_dict_lists(self):

        self.assertEqual(
            merge_dict(deepcopy(self.base_dict), deepcopy(self.delta_dict),
                       merge_lists=True, no_dupes=False),
            {"one": 1, "two": 2, "three": 30, "four": [
                "foo", "bar", "baz", 4, 5, 6, "foo"], "five": None}
        )

    def test_merge_dict_skip_empty(self):

        self.assertEqual(
            merge_dict(deepcopy(self.base_dict), deepcopy(self.delta_dict),
                       merge_lists=True, skip_empty=True, no_dupes=False),
            {"one": 1, "two": 2, "three": 30, "four": [
                "foo", "bar", "baz", 4, 5, 6, "foo"], "five": 50}
        )

        # NOT merging empty string / list / None
        self.assertEqual(
            merge_dict(deepcopy(self.base_dict),
                       {"val": None,  "val2": "",  "val3": []},
                       skip_empty=True), self.base_dict
        )

        # merging False and 0 and " "
        self.assertEqual(
            merge_dict(deepcopy(self.base_dict),
                       {"one": False, "two": 0, "four": " "},
                       skip_empty=True),
            {"one": False, "two": 0, "three": 3, "four": " ", "five": 50}
        )

    # Difficult to test because order is not currently guaranteed.
    def test_merge_dict_no_dupes(self):

        self.assertEqual(
            merge_dict(deepcopy(self.base_dict), deepcopy(self.delta_dict), merge_lists=True,
                       skip_empty=True, no_dupes=True),
            {"one": 1, "two": 2, "three": 30, "four": [
                "foo", "bar", "baz", 4, 5, 6], "five": 50}
        )



import os
from setuptools import setup

BASEDIR = os.path.abspath(os.path.dirname(__file__))


def required(requirements_file):
    """ Read requirements file and remove comments and empty lines. """
    with open(os.path.join(BASEDIR, requirements_file), 'r') as f:
        requirements = f.read().splitlines()
        if 'MYCROFT_LOOSE_REQUIREMENTS' in os.environ:
            print('USING LOOSE REQUIREMENTS!')
            requirements = [r.replace('==', '>=').replace('~=', '>=') for r in requirements]
        return [pkg for pkg in requirements
                if pkg.strip() and not pkg.startswith("#")]


setup(
    name='ovos_utils',
    version='0.0.14',
    packages=['ovos_utils',
              'ovos_utils.intents',
              'ovos_utils.sound',
              "ovos_utils.enclosure",
              'ovos_utils.enclosure.mark1',
              'ovos_utils.enclosure.mark1.eyes',
              'ovos_utils.enclosure.mark1.faceplate',
              'ovos_utils.skills',
              'ovos_utils.lang'],
    url='https://github.com/OpenVoiceOS/ovos_utils',
    install_requires=required("requirements/requirements.txt"),
    extras_require={
        "extras": required("requirements/extras.txt")
    },
    include_package_data=True,
    license='Apache',
    author='jarbasAI',
    author_email='jarbasai@mailfence.com',
    description='collection of simple utilities for use across the mycroft ecosystem'
)



import subprocess
import time
from ovos_utils.log import LOG
from ovos_utils.signal import check_for_signal
from distutils.spawn import find_executable


def play_audio(uri, play_cmd=None):
    """ Play a audio file.

        Returns: subprocess.Popen object
    """
    sox_play = find_executable("play")
    pulse_play = find_executable("paplay")
    alsa_play = find_executable("aplay")
    mpg123_play = find_executable("mpg123")

    player = play_cmd

    # NOTE: some urls like youtube streams will cause extension detection to fail
    # let's handle it explicitly
    ext = uri.split("?")[0].split(".")[-1]
    # Replace file:// uri's with normal paths
    uri = uri.replace('file://', '')

    # sox should handle almost every format, but fails in some urls
    if sox_play:
        player = sox_play + f" --type {ext} %1"
    # determine best available player
    else:
        # wav file
        if 'wav' in ext:
            if pulse_play:
                player = pulse_play + " %1"
            elif alsa_play:
                player = alsa_play + " %1"
        # guess mp3
        elif mpg123_play:
            player = mpg123_play + " %1"

    if not player:
        LOG.error(f"Failed to play: No playback functionality available")
        return None

    play_cmd = player.split(" ")

    for index, cmd in enumerate(play_cmd):
        if cmd == "%1":
            play_cmd[index] = uri

    try:
        return subprocess.Popen(play_cmd)
    except Exception as e:
        LOG.error(f"Failed to play: {play_cmd}")
        LOG.exception(e)
        return None


def play_wav(uri, play_cmd="paplay %1"):
    """ Play a wav-file.

        Returns: subprocess.Popen object
    """
    play_wav_cmd = str(play_cmd).split(" ")
    for index, cmd in enumerate(play_wav_cmd):
        if cmd == "%1":
            play_wav_cmd[index] = uri
    try:
        return subprocess.Popen(play_wav_cmd)
    except Exception as e:
        LOG.error("Failed to launch WAV: {}".format(play_wav_cmd))
        LOG.debug("Error: {}".format(repr(e)), exc_info=True)
        return None


def play_mp3(uri, play_cmd="mpg123 %1"):
    """ Play a mp3-file.

        Returns: subprocess.Popen object
    """
    play_mp3_cmd = str(play_cmd).split(" ")
    for index, cmd in enumerate(play_mp3_cmd):
        if cmd == "%1":
            play_mp3_cmd[index] = uri
    try:
        return subprocess.Popen(play_mp3_cmd)
    except Exception as e:
        LOG.error("Failed to launch MP3: {}".format(play_mp3_cmd))
        LOG.debug("Error: {}".format(repr(e)), exc_info=True)
        return None


def play_ogg(uri, play_cmd="ogg123 -q %1"):
    """ Play a ogg-file.

        Returns: subprocess.Popen object
    """
    play_ogg_cmd = str(play_cmd).split(" ")
    for index, cmd in enumerate(play_ogg_cmd):
        if cmd == "%1":
            play_ogg_cmd[index] = uri
    try:
        return subprocess.Popen(play_ogg_cmd)
    except Exception as e:
        LOG.error("Failed to launch OGG: {}".format(play_ogg_cmd))
        LOG.debug("Error: {}".format(repr(e)), exc_info=True)
        return None


def record(file_path, duration, rate, channels):
    if duration > 0:
        return subprocess.Popen(
            ["arecord", "-r", str(rate), "-c", str(channels), "-d",
             str(duration), file_path])
    else:
        return subprocess.Popen(
            ["arecord", "-r", str(rate), "-c", str(channels), file_path])


def is_speaking():
    """Determine if Text to Speech is occurring

    Returns:
        bool: True while still speaking
    """
    return check_for_signal("isSpeaking", -1)


def wait_while_speaking():
    """Pause as long as Text to Speech is still happening

    Pause while Text to Speech is still happening.  This always pauses
    briefly to ensure that any preceeding request to speak has time to
    begin.
    """
    time.sleep(0.3)  # Wait briefly in for any queued speech to begin
    while is_speaking():
        time.sleep(0.1)


import collections
import csv
import re
import os
from os import walk
from os.path import splitext, join, dirname
import tempfile
from ovos_utils.bracket_expansion import expand_options
from ovos_utils.log import LOG


def get_temp_path(*args):
    """Generate a valid path in the system temp directory.

    This method accepts one or more strings as arguments. The arguments are
    joined and returned as a complete path inside the systems temp directory.
    Importantly, this will not create any directories or files.

    Example usage: get_temp_path('mycroft', 'audio', 'example.wav')
    Will return the equivalent of: '/tmp/mycroft/audio/example.wav'

    Args:
        path_element (str): directories and/or filename

    Returns:
        (str) a valid path in the systems temp directory
    """
    try:
        path = os.path.join(tempfile.gettempdir(), *args)
    except TypeError:
        raise TypeError("Could not create a temp path, get_temp_path() only "
                        "accepts Strings")
    return path


def get_cache_directory(folder):
    # optional import to use ram for cache
    # does not work in windows!
    try:
        from memory_tempfile import MemoryTempfile
    except ImportError:
        MemoryTempfile = None
    if os.name == 'nt' or not MemoryTempfile:
        path = get_temp_path(folder)
    else:
        path = join(MemoryTempfile(fallback=True).gettempdir(), folder)
    os.makedirs(path, exist_ok=True)
    return path


def resolve_ovos_resource_file(res_name):
    """Convert a resource into an absolute filename.
    used internally for ovos resources
    """
    # First look for fully qualified file (e.g. a user setting)
    if os.path.isfile(res_name):
        return res_name

    # now look in bundled ovos resources
    filename = join(dirname(__file__), "res", res_name)
    if os.path.isfile(filename):
        return filename

    # let's look in ovos_workshop if it's installed
    try:
        import ovos_workshop
        pkg_dir = dirname(ovos_workshop.__file__)
        filename = join(pkg_dir, "res", res_name)
        if os.path.isfile(filename):
            return filename
        filename = join(pkg_dir, "res", "ui", res_name)
        if os.path.isfile(filename):
            return filename
    except:
        pass
    return None  # Resource cannot be resolved


def resolve_resource_file(res_name, root_path=None, config=None):
    """Convert a resource into an absolute filename.

    Resource names are in the form: 'filename.ext'
    or 'path/filename.ext'

    The system wil look for ~/.mycroft/res_name first, and
    if not found will look at /opt/mycroft/res_name,
    then finally it will look for res_name in the 'mycroft/res'
    folder of the source code package.

    Example:
    With mycroft running as the user 'bob', if you called
        resolve_resource_file('snd/beep.wav')
    it would return either '/home/bob/.mycroft/snd/beep.wav' or
    '/opt/mycroft/snd/beep.wav' or '.../mycroft/res/snd/beep.wav',
    where the '...' is replaced by the path where the package has
    been installed.

    Args:
        res_name (str): a resource path/name
        config (dict): mycroft.conf, to read data directory from
    Returns:
        str: path to resource or None if no resource found
    """
    if config is None:
        from ovos_utils.configuration import read_mycroft_config
        config = read_mycroft_config()

    # First look for fully qualified file (e.g. a user setting)
    if os.path.isfile(res_name):
        return res_name

    # Now look for ~/.mycroft/res_name (in user folder)
    filename = os.path.expanduser("~/.mycroft/" + res_name)
    if os.path.isfile(filename):
        return filename

    # Next look for /opt/mycroft/res/res_name
    data_dir = os.path.expanduser(config.get('data_dir', "/opt/mycroft"))
    filename = os.path.expanduser(os.path.join(data_dir, res_name))
    if os.path.isfile(filename):
        return filename

    # look in ovos_utils package itself
    found = resolve_ovos_resource_file(res_name)
    if found:
        return found

    # Finally look for it in the source package
    paths = [
        "/opt/venvs/mycroft-core/lib/python3.7/site-packages/",  # mark1/2
        "/opt/venvs/mycroft-core/lib/python3.4/site-packages/ ",
        # old mark1 installs
        "/home/pi/mycroft-core"  # picroft
    ]
    if root_path:
        paths += [root_path]
    for p in paths:
        filename = os.path.join(p, 'mycroft', 'res', res_name)
        filename = os.path.abspath(os.path.normpath(filename))
        if os.path.isfile(filename):
            return filename

    return None  # Resource cannot be resolved


def read_vocab_file(path):
    """ Read voc file.

        This reads a .voc file, stripping out empty lines comments and expand
        parentheses. It returns each line as a list of all expanded
        alternatives.

        Args:
            path (str): path to vocab file.

        Returns:
            List of Lists of strings.
    """
    vocab = []
    with open(path, 'r', encoding='utf8') as voc_file:
        for line in voc_file.readlines():
            if line.startswith('#') or line.strip() == '':
                continue
            vocab.append(expand_options(line.lower()))
    return vocab


def load_regex_from_file(path, skill_id):
    """Load regex from file
    The regex is sent to the intent handler using the message bus

    Args:
        path:       path to vocabulary file (*.voc)
        skill_id:   skill_id to the regex is tied to
    """
    from ovos_utils.intents.intent_service_interface import munge_regex

    regexes = []
    if path.endswith('.rx'):
        with open(path, 'r', encoding='utf8') as reg_file:
            for line in reg_file.readlines():
                if line.startswith("#"):
                    continue
                LOG.debug('regex pre-munge: ' + line.strip())
                regex = munge_regex(line.strip(), skill_id)
                LOG.debug('regex post-munge: ' + regex)
                # Raise error if regex can't be compiled
                try:
                    re.compile(regex)
                    regexes.append(regex)
                except Exception as e:
                    LOG.warning(f'Failed to compile regex {regex}: {e}')

    return regexes


def load_vocabulary(basedir, skill_id):
    """Load vocabulary from all files in the specified directory.

    Args:
        basedir (str): path of directory to load from (will recurse)
        skill_id: skill the data belongs to
    Returns:
        dict with intent_type as keys and list of list of lists as value.
    """
    from ovos_utils.intents.intent_service_interface import to_alnum

    vocabs = {}
    for path, _, files in walk(basedir):
        for f in files:
            if f.endswith(".voc"):
                vocab_type = to_alnum(skill_id) + splitext(f)[0]
                vocs = read_vocab_file(join(path, f))
                if vocs:
                    vocabs[vocab_type] = vocs
    return vocabs


def load_regex(basedir, skill_id):
    """Load regex from all files in the specified directory.

    Args:
        basedir (str): path of directory to load from
        bus (messagebus emitter): messagebus instance used to send the vocab to
                                  the intent service
        skill_id (str): skill identifier
    """
    regexes = []
    for path, _, files in walk(basedir):
        for f in files:
            if f.endswith(".rx"):
                regexes += load_regex_from_file(join(path, f), skill_id)
    return regexes


def read_value_file(filename, delim):
    """Read value file.

    The value file is a simple csv structure with a key and value.

    Args:
        filename (str): file to read
        delim (str): csv delimiter

    Returns:
        OrderedDict with results.
    """
    result = collections.OrderedDict()

    if filename:
        with open(filename) as f:
            reader = csv.reader(f, delimiter=delim)
            for row in reader:
                # skip blank or comment lines
                if not row or row[0].startswith("#"):
                    continue
                if len(row) != 2:
                    continue

                result[row[0]] = row[1]
    return result


def read_translated_file(filename, data):
    """Read a file inserting data.

    Args:
        filename (str): file to read
        data (dict): dictionary with data to insert into file

    Returns:
        list of lines.
    """
    if filename:
        with open(filename) as f:
            text = f.read().replace('{{', '{').replace('}}', '}')
            return text.format(**data or {}).rstrip('\n').split('\n')
    else:
        return None
